package ru.dbotthepony.mc.otm.android

import com.google.common.collect.ImmutableList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import net.minecraft.network.chat.Component
import net.minecraft.network.chat.ComponentContents
import net.minecraft.network.chat.MutableComponent
import net.minecraft.network.chat.contents.TranslatableContents
import ru.dbotthepony.mc.otm.TranslatableComponent
import ru.dbotthepony.mc.otm.capability.AndroidCapability
import ru.dbotthepony.mc.otm.getKeyNullable
import ru.dbotthepony.mc.otm.registry.MRegistry

fun interface AndroidResearchFactory<R : AndroidResearch> {
	fun factory(type: AndroidResearchType<*>, capability: AndroidCapability): R
}

private fun findPrerequisites(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
	top: Boolean = true
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		if (!top) {
			add.add(value)
		}

		findPrerequisites(value.definedPrerequisites, add, false)
	}

	return add
}

private fun findAllPrerequisites(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		add.add(value)
		findAllPrerequisites(value.flatPrerequisites, add)
	}

	return add
}

private fun findAllChildren(
	initial: Collection<AndroidResearchType<*>>,
	add: MutableSet<AndroidResearchType<*>> = HashSet(),
): Set<AndroidResearchType<*>> {
	for (value in initial) {
		add.add(value)
		findAllChildren(value.flatUnlocks, add)
	}

	return add
}

open class AndroidResearchType<R : AndroidResearch>(
	protected val factory: AndroidResearchFactory<R>
) {
	val researchTreeDepth: Int by lazy {
		if (flatPrerequisites.isEmpty()) {
			return@lazy 0
		}

		var depth = 1

		for (value in flatPrerequisites) {
			depth = depth.coerceAtLeast(value.researchTreeDepth + 1)
		}

		return@lazy depth
	}

	/**
	 * Prerequisites as-is
	 */
	open val definedPrerequisites: List<AndroidResearchType<*>> get() = emptyList()

	/**
	 * Blocked by list as-is
	 */
	open val definedBlockedBy: List<AndroidResearchType<*>> get() = emptyList()

	/**
	 * Flat list of research preceding this research.
	 *
	 * This list won't contain research following the same research path twice (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B depends on A
	 *
	 * C specify both B and A as it's prerequisites, [flatPrerequisites] will contain only B, when [definedPrerequisites] will contain
	 * both B and A
	 */
	val flatPrerequisites: List<AndroidResearchType<*>> by lazy {
		val parentPrerequisites = findPrerequisites(definedPrerequisites)
		val builder = ImmutableList.builder<AndroidResearchType<*>>()

		for (value in definedPrerequisites) {
			if (value !in parentPrerequisites) {
				builder.add(value)
			}
		}

		builder.build()
	}

	/**
	 * All prerequisite research, including indirect ones
	 */
	val allPrerequisites: List<AndroidResearchType<*>> by lazy {
		ImmutableList.copyOf(findAllPrerequisites(flatPrerequisites))
	}

	/**
	 * Flat list of research blocked by this research.
	 *
	 * This list won't contain children of any blocked research (guaranteed to be flat).
	 *
	 * E.g.
	 *
	 * * C depends on B
	 * * B is blocked by A
	 * * C is blocked by A
	 *
	 * Both C and B specify A as it's blockers, [flatBlocking] will contain only B, because C is blocked by A through B.
	 */
	val flatBlocking: List<AndroidResearchType<*>> by lazy {
		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (this in research.definedBlockedBy) {
				var hit = false

				for (parent in research.allPrerequisites) {
					if (this in parent.definedBlockedBy) {
						hit = true
						break
					}
				}

				if (!hit) {
					list.add(research)
				}
			}
		}

		return@lazy list.build()
	}

	/**
	 * All research directly unlocked by this research.
	 */
	val flatUnlocks: List<AndroidResearchType<*>> by lazy {
		val list = ImmutableList.builder<AndroidResearchType<*>>()

		for (research in MRegistry.ANDROID_RESEARCH) {
			if (this in research.flatPrerequisites) {
				list.add(research)
			}
		}

		return@lazy list.build()
	}

	/**
	 * All research unlocked by this research, including indirect ones
	 */
	val allUnlocks: List<AndroidResearchType<*>> by lazy {
		ImmutableList.copyOf(findAllChildren(flatUnlocks))
	}

	fun factory(capability: AndroidCapability) = factory.factory(this, capability)

	val registryName by lazy {
		MRegistry.ANDROID_RESEARCH.getKeyNullable(this)
	}

	val displayId by lazy {
		val registryName = registryName ?: throw NullPointerException("No registry name present")
		return@lazy "android_research.${registryName.namespace}.${registryName.path}".intern()
	}

	val descriptionId by lazy {
		return@lazy "$displayId.description".intern()
	}

	open val displayContents: ComponentContents by lazy {
		TranslatableContents(displayId)
	}

	open val displayName: Component by lazy {
		MutableComponent.create(displayContents)
	}

	open val displayDescription: List<Component> by lazy {
		listOf(TranslatableComponent(descriptionId))
	}
}
